/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_MONEYPARTS_H
#define PECUNIA_MONEYPARTS_H

#include <cstdint>
#include <optional>

#ifdef __WIN32
#	pragma STDC FENV_ACCESS ON
#endif

#include <pecunia/pecunia_export.h>
#include <pecunia/Sizes.hpp>


namespace pecunia
{

class FloatingPoint;

}
namespace pecunia::currency
{

enum class Currency : std::uint16_t;
class Money;

}
namespace pecunia
{

/**
 * @brief A representation of all parts of a monetary value.
 */
class MoneyParts
{
	// Justification: Used for performance reasons.
	friend class pecunia::currency::Money;

	MoneyParts(
		const MajorUnit full,
		const MajorUnit major,
		const MinorUnit minorWithSub,
		const std::optional<MinorUnit> minorOnly,
		const MinorUnit subMinor,
		const pecunia::currency::Currency code
	);

	MajorUnit full_;
	MajorUnit major_;
	MinorUnit minorWithSub_;
	std::optional<MinorUnit> minorOnly_;
	MinorUnit subMinor_;
	pecunia::currency::Currency code_;

public:
	PECUNIA_EXPORT MoneyParts();
	/**
	 * @brief Component constructor that fully initialises the object.
	 *
	 * @exception InvalidArgument When the monetary value does not support minor units and minor
	 * units are supplied.
	 *
	 * @param code The currency of the monetary value.
	 * @param major The major monetary value.
	 * @param minorOnly The minor monetary value, if the currency supports it.
	 * @param subMinor The sub-minor monetary value.
	 */
	PECUNIA_EXPORT MoneyParts(
		const pecunia::currency::Currency code,
		const MajorUnit major,
		const std::optional<MinorUnit> minorOnly,
		const MinorUnit subMinor = 0
	);
	PECUNIA_EXPORT MoneyParts(const MoneyParts&) = default;
	PECUNIA_EXPORT MoneyParts(MoneyParts&&) = default;
	PECUNIA_EXPORT MoneyParts& operator=(const MoneyParts&) = default;
	PECUNIA_EXPORT MoneyParts& operator=(MoneyParts&&) = default;

	/// An accessor to the full amount monetary value.
	[[nodiscard]]
	PECUNIA_EXPORT const MajorUnit& full() const noexcept;
	/// An accessor to the major monetary value.
	[[nodiscard]]
	PECUNIA_EXPORT const MajorUnit& major() const noexcept;
	/// An accessor to the minor and sub-minor monetary value.
	[[nodiscard]]
	PECUNIA_EXPORT const MinorUnit& minorWithSub() const noexcept;
	/// An accessor to the minor monetary value.
	[[nodiscard]]
	PECUNIA_EXPORT const std::optional<MinorUnit>& minorOnly() const noexcept;
	/// An accessor to the sub-minor monetary value.
	[[nodiscard]]
	PECUNIA_EXPORT const MinorUnit& subMinor() const noexcept;
	/// An accessor to the ISO currency code of the monetary amount.
	[[nodiscard]]
	PECUNIA_EXPORT const pecunia::currency::Currency& code() const noexcept;
};

/**
 * @brief General equality operator.
 *
 * @param lhs The object that is expected to be compared.
 * @param rhs The object that is to compare for equality.
 *
 * @return True when all member variables are the same, false otherwise.
 */
[[nodiscard]]
PECUNIA_EXPORT bool operator==(const MoneyParts& lhs, const MoneyParts& rhs);

}

#endif
